(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umlmon_disk.ml 187 2005-10-14 10:54:15Z gerd $ *)

open Umlmonctrl_aux
open Umlmon_types
open Umlmon_common
open Cow_aux


let array_mem x a =
  let rec search k =
    if k >= Array.length a then
      raise Not_found
    else
      a.(k) = x || search (k+1)
  in
  search 0
;;


let disk_access global spec (Jailfile file) =
  let name = Filename.concat spec.umljaildir file in
  let (r,w,x) = file_access name spec.umlpriv in
  
  if r && w then
    `Read_write
  else
    if r then
      `Read_only
    else
      `Invisible
;;


let read_cow_header filename =
  let l = Rtypes.int_of_uint4 header_len_v3 in
  let buf = String.create l in
  let f = open_in filename in
  let h =
    ( try
	really_input f buf 0 l;
	let xt = Xdr.validate_xdr_type xdrt_cow_header_v3 in
	let xv = Xdr.unpack_xdr_value ~fast:true buf xt [] in
	_to_cow_header_v3 xv
      with
	| End_of_file -> close_in f; failwith ("File is too short: " ^ filename)
	| err -> close_in f; raise err
    ) in
  close_in f;
  if h.chv3_magic <> (Rtypes.int32_of_uint4 cow_magic) then
    failwith ("File has wrong magic number: " ^ filename);
  if h.chv3_version <> 3l then
    failwith ("Not a version 3 .cow file: " ^ filename);
  h
;;


let extract_backing_file cow_header =
  try
    let n = String.index cow_header.chv3_backing_file '\000' in
    String.sub cow_header.chv3_backing_file 0 n
  with
      Not_found ->
	failwith ("Illegal .cow header")
;;


let write_cow_header fd h =
  let xt = Xdr.validate_xdr_type xdrt_cow_header_v3 in
  let xv = _of_cow_header_v3 h in
  let s = Xdr.pack_xdr_value_as_string xv xt [] in
  ignore(Unix.write fd s 0 (String.length s))
;;


let create_cow_file gs bak_filename cow_filename =
  let spec = Lazy.force gs.umlspec in
  let jd = spec.umljaildir in
  let uid = spec.umlpriv.priv_uid in
  let gid = spec.umlpriv.priv_gid in
  let bak_length = Rtypes.int_of_uint4 path_len_v3 in

  if String.length bak_filename >= bak_length then
    failwith "File name of backing file is too long";

  let bak_filename_padded = String.make bak_length '\000' in
  String.blit bak_filename 0 bak_filename_padded 0 (String.length bak_filename);

  let full_bak_filename = Filename.concat jd bak_filename in
  let st = Unix.LargeFile.stat full_bak_filename in

  let sectorsize = 512 in
  let alignment = Linux.Sysconf.pagesize() in        (* i.e. sysconf(_SC_PAGESIZE) *)
  let h =
    { chv3_magic = Rtypes.int32_of_uint4 cow_magic;
      chv3_version = 3l;
      chv3_mtime = Int32.of_float(st.Unix.LargeFile.st_mtime);
      chv3_size = st.Unix.LargeFile.st_size;
      chv3_sectorsize = Int32.of_int sectorsize;
      chv3_alignment = Int32.of_int alignment;
      chv3_cow_format = 0l;
      chv3_backing_file = bak_filename_padded;
    } in

  let full_cow_filename = Filename.concat jd cow_filename in

  let fd = 
    Unix.openfile full_cow_filename [Unix.O_RDWR; Unix.O_CREAT; Unix.O_EXCL] 0o600 in
  Unix.fchown fd uid gid;

  let div_round x len = (x+len-1) / len in
  let div_round_64 x len = Int64.div (Int64.pred (Int64.add x len)) len in
  let round_up x align = (div_round x align) * align in

  try
    write_cow_header fd h;
    
    let size = st.Unix.LargeFile.st_size in
    let bitmap_offset = round_up (Rtypes.int_of_uint4 header_len_v3) alignment in
    let bitmap_len_64 = div_round_64 (div_round_64 size (Int64.of_int sectorsize)) 8L in
    let bitmap_len = Int64.to_int bitmap_len_64 in
    let data_offset = round_up (bitmap_offset + bitmap_len) alignment in

    let eof_pos = Int64.add (Int64.of_int data_offset) size in

    ignore(Unix.LargeFile.lseek fd (Int64.pred eof_pos) Unix.SEEK_SET);
    ignore(Unix.write fd "\000" 0 1);

    Unix.close fd

  with
      err -> Unix.close fd; raise err
;;


let locations = [ "/disks"; "/shared/disks" ]

let owner_re = Str.regexp "^/disks/" 

let read_disks gs =
  let global = Lazy.force gs.umlglobal in
  let spec = Lazy.force gs.umlspec in
  let jd = spec.umljaildir in
  let shared_locks = Umlmon_shared.get_locks global spec in

  let files =
    List.flatten
      (List.map
	 (fun loc ->
	    let d = Filename.concat jd loc in
	    if Sys.file_exists d then
	      List.map 
		(fun f -> Filename.concat loc f)
		(Array.to_list (Sys.readdir d))
	    else
	      []
	 )
	 locations
      ) in

  let files' =
    List.filter
      (fun f ->
	 ( Filename.check_suffix f ".cow" || 
	   Filename.check_suffix f ".dsk" ||
	   Filename.check_suffix f ".dev"
	 ) &&
	   disk_access global spec (Jailfile f) <> `Invisible
      )
      files in

  let get_type f =
    if Filename.check_suffix f ".cow" then (
      let h = read_cow_header (Filename.concat jd f) in
      `Cow { cow_backing_file = Jailfile (extract_backing_file h) }
    )
    else
      if Filename.check_suffix f ".dsk" then
	`Dsk
      else
	if Filename.check_suffix f ".dev" then
	  `Dev
	else
	  assert false in

  let is_owned f =
    Str.string_match owner_re f 0 in

  let get_lock_status name =
    try
      let (lock, _, our_lock) =
	List.find
	  (fun (_, n, _) -> n = Jailfile name)
	  shared_locks in
      match (lock, our_lock) with
	| (`Read, _) -> 
	    let have_lock =
	      List.exists
		(fun (_, n, o) -> o && n = Jailfile name) shared_locks in
	    if have_lock then `Read_own else `Read
	| (`Write, false) -> `Write
	| (`Write, true) -> `Write_own
    with
	Not_found -> `None 
  in

  List.map
    (fun name ->
       { md_file_name = Jailfile name;
	 md_file_type = get_type name;
	 md_file_owned = is_owned name;
	 md_file_access = disk_access global spec (Jailfile name);
	 md_file_status = `Exists;
	 md_file_lock = get_lock_status name
       }
    )
    files'
;;


let get_disks gs =
  match gs.md with
    | None -> 
	let disks = read_disks gs in
	gs.md <- Some disks;
	disks
    | Some disks ->
	disks
;;


let operation_in_progress gs =
  match gs.md with
    | None -> false
    | Some disks ->
	List.exists (fun disk -> disk.md_file_status = `Being_created) disks


let use_count disks filename =
  (* Count how many cow files refer to [filename] *)
  List.fold_left
    (fun acc disk ->
       match disk.md_file_type with
	 | `Cow n ->
	     if n.cow_backing_file = filename then acc+1 else acc
	 | _ ->
	     acc
    )
    0
    disks
;;


let lock_disks_at_startup gs =
  let global = Lazy.force gs.umlglobal in
  let spec = Lazy.force gs.umlspec in
  let disknames = List.map (fun ud -> ud.diskname) spec.umldisks in
  if has_dups disknames then
    failwith "Duplicate disk devices";
  let disks = get_disks gs in

  (* Check the sanity of the disk definition, and collect the locking
   * requirements: 
   *)
  let locks = ref [] in
  List.iter
    (fun ud ->
       let file = ud.diskdef in
       let Jailfile jf = file in
       (* Check whether [file] is known, and its type: *)
       try
	 let md = List.find (fun md -> md.md_file_name = file) disks in
	 if md.md_file_access <> `Read_write then
	   failwith ("File not writeable: " ^ jf);
	 if not md.md_file_owned then
	   locks := (`Write, file) :: !locks;
	 match md.md_file_type with
	   | `Cow info ->
	       let bfile = info.cow_backing_file in
	       let Jailfile bjf = bfile in
	       (* Check [bfile]: *)
	       ( try
		   let bmd = List.find (fun bmd -> bmd.md_file_name = bfile) disks in
		   ( match bmd.md_file_type with
		       | `Cow _ ->
			   failwith ("Backing file " ^ bjf ^ " is again a cow file")
		       | `Dsk | `Dev ->
			   ()
		   );
		   locks := (`Read, bfile) :: !locks
		 with
		     Not_found ->
		       failwith ("Backing file not found: " ^ bjf ^ 
				 " - for device " ^ ud.diskname)
	       )
	   | `Dsk
	   | `Dev ->
	       if use_count disks file > 0 then
		 failwith ("File is in use as backing file: " ^ jf ^ 
			   " - for device " ^ ud.diskname);
       with
	   Not_found ->
	     failwith ("Disk file not found: " ^ jf ^ " - for device " ^ ud.diskname)
    )
    spec.umldisks;

  (* Now lock the shared devices: *)
  Umlmon_shared.lock_disks global spec !locks
;;


let unlock_disks gs () =
  let global = Lazy.force gs.umlglobal in
  let spec = Lazy.force gs.umlspec in
  Umlmon_shared.unlock_all_disks global spec
;;


let kernel_arguments gs =
  (* Note: Sanity checks now in lock_disks! *)
  let spec = Lazy.force gs.umlspec in
  let disks = get_disks gs in
  List.map
    (fun ud ->
       let file = ud.diskdef in
       let Jailfile jf = file in
       (* Check whether [file] is known, and its type: *)
       try
	 let md = List.find (fun md -> md.md_file_name = file) disks in
	 match md.md_file_type with
	   | `Cow info ->
	       let bfile = info.cow_backing_file in
	       let Jailfile bjf = bfile in
	       (* Check [bfile]: *)
	       let bmd = List.find (fun bmd -> bmd.md_file_name = bfile) disks in
	       ud.diskname ^ "s=" ^ jf ^ "," ^ bjf
	   | `Dsk
	   | `Dev ->
	       ud.diskname ^ "s=" ^ jf
       with
	   Not_found ->
	     failwith "Cannot determine kernel arguments for disks"
    )
    spec.umldisks


let mk_lock_info =
  function
    | `None -> `lock_null
    | `Read -> `lock_read
    | `Read_own -> `lock_read_own
    | `Write -> `lock_write
    | `Write_own -> `lock_write_own
;;


let proc_disk_files ues gs =
  protect
    (fun session _ reply ->
       Umlmon_auth.ensure_authenticated session;
       let spec = Lazy.force gs.umlspec in
       let jd = spec.umljaildir in
       let disks = get_disks gs in
       let disks' =
	 List.filter
	   (fun disk -> 
	      let Jailfile jf = disk.md_file_name in
	      Sys.file_exists (Filename.concat jd jf))
	   disks in
       let r_disks =
	 List.map
	   (fun d ->
	      let Jailfile djf = d.md_file_name in
	      let disk_info =
		match d.md_file_status with
		  | `Exists ->
		      let size = 
			(* The following method works for both regular files
                         * and block devices:
                         *)
			( let fd = 
			    Unix.openfile 
			      (Filename.concat jd djf) [ Unix.O_RDONLY ] 0 in
			  try
			    let s = Unix.LargeFile.lseek fd 0L Unix.SEEK_END in
			    Unix.close fd;
			    s
			  with
			    | error -> Unix.close fd; raise error
			) in
		      `disk_exists 
			{ disk_size = size;
			  disk_allocated = 0L;
			    (* TODO. This _is_ returned by orig Linux stat *)
			  disk_backing_file = 
			    ( match d.md_file_type with
				| `Cow n -> 
				    let Jailfile njf = n.cow_backing_file in
				    Some njf
				| `Dsk | `Dev   -> None );
			  disk_use_count = Int32.of_int (use_count disks d.md_file_name);
			  disk_rw = (d.md_file_access = `Read_write);
			  disk_lock = mk_lock_info d.md_file_lock;
			  disk_extension = [| |];
			}
		  | `Being_created ->
		      `disk_being_created
		  | `Error msg ->
		      `disk_error msg
	      in
	      { disk_name = djf;
		disk_is_cow = 
		  ( match d.md_file_type with
		      | `Cow _ -> true | _ -> false );
		disk_owned = d.md_file_owned;
		disk_info = disk_info;
	      }
	   )
	   disks' in

       (* Now reset all `Error entries - they have been reported *)
       List.iter
	 (fun disk ->
	    match disk.md_file_status with
	      | `Error _ -> 
		  disk.md_file_status <- `Exists
	      | _ -> 
		  ()
	 )
	 disks;

       (* If all disks have status `Exists, force reloading the disk data: *)
       if List.for_all (fun disk -> disk.md_file_status = `Exists) disks then
	 gs.md <- None;

       reply (Array.of_list r_disks)
    )
;;


let md_enter_disk gs name new_status new_type new_owned =
  let disks = get_disks gs in
  let disks' =
    if List.exists (fun disk -> disk.md_file_name = name) disks then
      List.map
	(fun disk ->
	   if disk.md_file_name = name then
	     { disk with 
		 md_file_status = new_status;
		 md_file_type = new_type;
		 md_file_owned = new_owned
	     }
	   else
	     disk
	)
	disks
    else
      disks @ 
      [ { md_file_name = name;
	  md_file_type = new_type;
	  md_file_status = new_status;
	  md_file_owned = new_owned;
	  md_file_access = `Read_write;
	  md_file_lock = `None;
	} ]
  in
  gs.md <- Some disks'
;;  


let md_remove_disk gs name =
  let disks = get_disks gs in
  let disks' =
    List.filter
      (fun disk -> disk.md_file_name <> name)
      disks in
  gs.md <- Some disks'
;;  



class zero_channel n : Uq_engines.async_in_channel =
object (self)
  val mutable notify_list = []
  val mutable notify_list_new = []

  val mutable pos = 0L
  val mutable closed = false

  method input s p l =
    let m = Int64.to_int (min (Int64.of_int l) (Int64.sub n pos)) in
    if l > 0 && m = 0 then raise End_of_file;
    String.fill s p m '\000';
    pos <- Int64.add pos (Int64.of_int m);
    m

  method close_in() = closed <- true; self # notify()

  method pos_in = Int64.to_int pos

  method can_input = not closed

  method request_notification f =
    notify_list_new <- f :: notify_list_new

  method private notify() =
    notify_list <- notify_list @ notify_list_new;
    notify_list_new <- [];
    notify_list <- List.filter (fun f -> f()) notify_list

end
;;


let setup_write_zeros ues gs name fd n new_type new_owned =
  let eng = new Uq_engines.sender ~src:(new zero_channel n) ~dst:fd ~close_dst:true ues in
  Uq_engines.when_state
    ~is_done:(fun _ ->
		md_enter_disk gs name `Exists new_type new_owned)
    ~is_error:(fun e ->
		 let msg = Printexc.to_string e in
		 md_enter_disk gs name (`Error msg) new_type new_owned)
    eng
;;


let dsk_name_re = Str.regexp "/disks/[^/]+\\.dsk$"

let proc_disk_create ues gs =
  protect
    (fun session (filename,size,options) reply ->
       Umlmon_auth.ensure_authenticated session;
       let spec = Lazy.force gs.umlspec in
       let jd = spec.umljaildir in
       let uid = spec.umlpriv.priv_uid in
       let gid = spec.umlpriv.priv_gid in
       let sparse = array_mem "sparse" options in
       try
	 if gs.umlvm <> `Inactive then 
	   failwith "Disk operations are not permitted when the VM is active";
	 if not (Str.string_match dsk_name_re filename 0) then
	   failwith "Bad disk name - must be /disks/*.dsk";
	 if size <= 0l then
	   failwith "Invalid size";
	 let fullname = Filename.concat jd filename in
	 if Sys.file_exists fullname then
	   failwith "Disk exists already";
	 let fd = 
	   try Unix.openfile fullname [Unix.O_RDWR; Unix.O_CREAT; Unix.O_EXCL] 0o600
	   with
	       Unix.Unix_error(_,_,_) as e ->
		 failwith(Umlmon_common.string_of_unix_error e) in
	 let size_bytes = Int64.mul (Int64.of_int32 size) 1048576L in
	 ( try
	     Unix.fchown fd uid gid;
	     if sparse then (
	       ignore(Unix.LargeFile.lseek fd (Int64.pred size_bytes) Unix.SEEK_SET);
	       ignore(Unix.write fd " " 0 1);
	       Unix.close fd;
	       md_enter_disk gs (Jailfile filename) `Exists `Dsk true;
	       reply `code_ok
	     )
	     else (
	       setup_write_zeros 
		 ues gs (Jailfile filename) fd size_bytes `Dsk true;
	       md_enter_disk gs (Jailfile filename) `Being_created `Dsk true;
	       reply `code_in_progress
	     )
	   with
	       err -> Unix.close fd; raise err
	 )
       with
	 | Failure msg ->
	     reply (`code_error msg)
    )
;;


let proc_disk_resize ues gs =
  protect
    (fun session (filename,size,options) reply ->
       Umlmon_auth.ensure_authenticated session;
       let spec = Lazy.force gs.umlspec in
       let jd = spec.umljaildir in
       let disks = get_disks gs in
       let sparse = array_mem "sparse" options in
       try
	 if gs.umlvm <> `Inactive then 
	   failwith "Disk operations are not permitted when the VM is active";
	 if not (Str.string_match dsk_name_re filename 0) then
	   failwith "Bad disk name - must be /disks/*.dsk";
	 if size <= 0l then
	   failwith "Invalid size";
	 let fullname = Filename.concat jd filename in
	 if not (Sys.file_exists fullname) then
	   failwith "Disk does not exist";
	 if List.exists (fun disk -> 
			   disk.md_file_name = Jailfile filename &&
		            (disk.md_file_status = `Being_created ||
 			    use_count disks (Jailfile filename) > 0))
	                disks
	 then
	   failwith "Disk is busy";
	 let fd = Unix.openfile fullname [Unix.O_RDWR ] 0 in
	 let size_bytes = Int64.mul (Int64.of_int32 size) 1048576L in
	 ( try
	     let st = Unix.LargeFile.fstat fd in
	     if size_bytes < st.Unix.LargeFile.st_size then (
	       (* Shrink *)
	       Unix.LargeFile.ftruncate fd size_bytes;
	       Unix.close fd;
	       md_enter_disk gs (Jailfile filename) `Exists `Dsk true;
	       reply `code_ok
	     )
	     else (
	       (* Expand *)
	       if sparse then (
		 ignore(Unix.LargeFile.lseek fd (Int64.pred size_bytes) Unix.SEEK_SET);
		 ignore(Unix.write fd " " 0 1);
		 Unix.close fd;
		 md_enter_disk gs (Jailfile filename) `Exists `Dsk true;
		 reply `code_ok
	       )
	       else (
		 let old_size = st.Unix.LargeFile.st_size in
		 let new_bytes = Int64.sub size_bytes old_size in
		 ignore(Unix.LargeFile.lseek fd old_size Unix.SEEK_SET);
		 setup_write_zeros 
		   ues gs (Jailfile filename) fd new_bytes `Dsk true;
		 md_enter_disk gs (Jailfile filename) `Being_created `Dsk true;
		 reply `code_in_progress
	       )
	     )
	   with
	       err -> Unix.close fd; raise err
	 )
       with
	 | Failure msg ->
	     reply (`code_error msg)
	 | Unix.Unix_error(_,_,_) as e ->
	     reply (`code_error (Umlmon_common.string_of_unix_error e))
    )
;;


let src_name_re = Str.regexp "\\(/shared\\)?/disks/[^/]+\\.dsk$"

let cow_name_re = Str.regexp "/disks/[^/]+\\.cow$"

let proc_disk_create_cow ues gs =
  protect
    (fun session (bak_filename,cow_filename,options) reply ->
       Umlmon_auth.ensure_authenticated session;
       let spec = Lazy.force gs.umlspec in
       let jd = spec.umljaildir in
       let uid = spec.umlpriv.priv_uid in
       let gid = spec.umlpriv.priv_gid in
       try
	 if gs.umlvm <> `Inactive then 
	   failwith "Disk operations are not permitted when the VM is active";
	 if not (Str.string_match src_name_re bak_filename 0) then
	   failwith "Bad disk name - must be /disks/*.dsk or /shared/disks/*.dsk";
	 if not (Str.string_match cow_name_re cow_filename 0) then
	   failwith "Bad cow name - must be /disks/*.cow";
	 let bak_fullname = Filename.concat jd bak_filename in
	 let cow_fullname = Filename.concat jd cow_filename in
	 if not (Sys.file_exists bak_fullname) then
	   failwith "Backing disk does not exist";
	 if Sys.file_exists cow_fullname then
	   failwith "Disk exists already";
	 create_cow_file gs bak_filename cow_filename;
	 md_enter_disk gs (Jailfile cow_filename) `Exists 
	   (`Cow { cow_backing_file = Jailfile bak_filename }) true;
	 reply `code_ok
       with
	 | Failure msg ->
	     reply (`code_error msg)
	 | Unix.Unix_error(_,_,_) as e ->
	     reply (`code_error (Umlmon_common.string_of_unix_error e))
    )
;;


let own_name_re = Str.regexp "/disks/[^/]+\\.\\(cow\\|dsk\\)$"

let proc_disk_delete ues gs =
  protect
    (fun session (filename,options) reply ->
       Umlmon_auth.ensure_authenticated session;
       let spec = Lazy.force gs.umlspec in
       let jd = spec.umljaildir in
       let disks = get_disks gs in
       try
	 if gs.umlvm <> `Inactive then 
	   failwith "Disk operations are not permitted when the VM is active";
	 if not (Str.string_match own_name_re filename 0) then
	   failwith "Bad disk name - must be /disks/*.dsk or /disks/*.cow";
	 let fullname = Filename.concat jd filename in
	 if not (Sys.file_exists fullname) then
	   failwith "Disk does not exist";
	 if List.exists (fun disk -> 
			   disk.md_file_name = Jailfile filename &&
		            (disk.md_file_status = `Being_created ||
 			    use_count disks (Jailfile filename) > 0))
	                disks
	 then
	   failwith "Disk is busy";
	 Unix.unlink fullname;
	 md_remove_disk gs (Jailfile filename);
	 reply `code_ok
       with
	 | Failure msg ->
	     reply (`code_error msg)
	 | Unix.Unix_error(_,_,_) as e ->
	     reply (`code_error (Umlmon_common.string_of_unix_error e))
    )
;;


let cpsrc_name_re = Str.regexp "\\(/shared\\)?/disks/[^/]+\\.\\(dsk\\|cow\\|dev\\)$"

let cpdst_name_re = Str.regexp "/disks/[^/]+\\.\\(cow\\|dsk\\)$"
	 
let extension_re = Str.regexp ".*\\.\\(cow\\|dsk\\|dev\\)$"
  
let get_extension s =
  if Str.string_match extension_re s 0 then
    Str.matched_group 1 s
  else
    ""

let proc_disk_copy ues gs =
  protect
    (fun session (src_filename,dst_filename,options) reply ->
       Umlmon_auth.ensure_authenticated session;
       let global = Lazy.force gs.umlglobal in
       let spec = Lazy.force gs.umlspec in
       let jd = spec.umljaildir in
       let disks = get_disks gs in
       let uid = spec.umlpriv.priv_uid in
       let gid = spec.umlpriv.priv_gid in

       let to_close = ref [] in

       try
	 if gs.umlvm <> `Inactive then 
	   failwith "Disk operations are not permitted when the VM is active";
	 if not (Str.string_match cpsrc_name_re src_filename 0) then
	   failwith "Bad source disk name - must be /disks/*.{dsk,cow,dev} or /shared/disks/*.{dsk,cow,dev}";
	 if not (Str.string_match cpdst_name_re dst_filename 0) then
	   failwith "Bad destination disk name - must be /disks/*.{dsk,cow}";
	 
	 let src_ext = get_extension src_filename in
	 let dst_ext = get_extension dst_filename in
	 let src_is_cow = src_ext = ".cow" in
	 let dst_is_cow = dst_ext = ".cow" in
	 if src_is_cow <> dst_is_cow then
	   failwith "Source and destination file must have the same types";

	 let src_fullname = Filename.concat jd src_filename in
	 let dst_fullname = Filename.concat jd dst_filename in

	 if not (Sys.file_exists src_fullname) then
	   failwith "Source disk does not exist";
	 let src_disk =
	   try
	     List.find 
	       (fun disk -> disk.md_file_name = Jailfile src_filename)
	       disks
	   with 
	       Not_found -> failwith "Source disk not found" in

	 if src_disk.md_file_status = `Being_created then
	   failwith "Source disk is busy";
	 (* It is harmless when there are references to the source disk! *)
	 
	 if Sys.file_exists dst_fullname then
	   failwith "Destination disk already exists";

	 let src_fd = Unix.openfile src_fullname [Unix.O_RDONLY] 0 in
	 to_close := src_fd :: !to_close;

	 let dst_fd = 
	   Unix.openfile dst_fullname [Unix.O_WRONLY; Unix.O_CREAT; Unix.O_TRUNC] 0o600 in
	 to_close := dst_fd :: !to_close;

	 Unix.fchown dst_fd uid gid;

	 (* Lock the source file, so it cannot be modified in the meantime: *)
	 let lock_id =
	   if src_disk.md_file_owned then
	     None   (* Do not lock our own files *)
	   else
	     Some(Umlmon_shared.lock_disks
		    global spec [`Read, Jailfile src_filename]) in
	 (* May fail with [Lock_failure] *)
	 (* CHECK: What happens with the lock if something fails now? *)

	 let src_type = src_disk.md_file_type in

	 md_enter_disk gs (Jailfile dst_filename) `Being_created src_type true;

	 let eng =
	   new Uq_engines.copier (`Unidirectional(src_fd, dst_fd)) ues in
	 
	 let unlock() =
	   match lock_id with
	     | None -> ()
	     | Some id -> Umlmon_shared.unlock_disks global spec id in

	 Uq_engines.when_state
	   ~is_done:(fun _ -> 
		       unlock();
		       md_enter_disk
			 gs (Jailfile dst_filename) `Exists src_type true)
	   ~is_error:(fun e ->
			unlock();
			let msg = Printexc.to_string e in
			md_enter_disk 
			  gs (Jailfile dst_filename) (`Error msg) src_type true)
	   eng;

	 reply `code_in_progress
       with
	 | Failure msg ->
	     List.iter Unix.close !to_close;
	     reply (`code_error msg)
	 | Lock_failure msg ->
	     List.iter Unix.close !to_close;
	     reply (`code_error ("Lock failure: " ^ msg))
	 | Unix.Unix_error(_,_,_) as e ->
	     List.iter Unix.close !to_close;
	     reply (`code_error (Umlmon_common.string_of_unix_error e))
	 | err ->
	     List.iter Unix.close !to_close;
	     raise err
    )
;;

	   
let proc_disk_options ues gs =
  protect
    (fun session _ reply ->
       Umlmon_auth.ensure_authenticated session;
       reply
	 { disk_opts_create = [| "sparse" |];
	   disk_opts_create_cow = [| |];
	   disk_opts_delete = [| |];
	   disk_opts_copy = [| |];
	   disk_opts_resize = [| "sparse" |];
	 }
    )
;;
